/*

*/

import renderer.scene.*;
import renderer.models.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;
import java.io.File;
import java.util.List;
import java.util.ArrayList;

/**

*/
public class InteractiveGRSModels_R5 implements KeyListener, ComponentListener
{
   private final Scene scene; // The event handlers need access to these fields.
   private final List<Model> modelArray = new ArrayList<>();
   private final JFrame jf;
   private final FrameBufferPanel fbp;
   private int currentModel = 0;
   private double deltaX = 0.0;
   private double deltaY = 0.0;
   private double deltaS = 1.0;
   private boolean displayTransformations = false;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public InteractiveGRSModels_R5()
   {
      // Create the Scene object that we shall render.
      scene = new Scene();
      scene.camera.perspective = false;

      // Instantiate all the grs models.
      modelArray.add(new GRSModel(new File("assets/grs/bronto.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/rex.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/bronto_v2.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/rex_v2.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/dragon.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/vinci.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/usa.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/usa_v2.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/birdhead.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/knight.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/house.grs")));
      modelArray.add(new GRSModel(new File("assets/grs/scene.grs")));
      // Create a set of x and y axes.
      final Model axes = new Axes2D(-1, +1, -1, +1, 20, 20);

      // Give each model a random color.
      for (final Model m : modelArray)
      {
         ModelShading.setRandomColor(m);
      }
      ModelShading.setColor(axes, Color.red);

      // Add two models to the Scene.
      scene.addModel(modelArray.get(0),
                     axes);


      // Define initial dimensions for a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;

      // Create a FrameBufferPanel that holds a FrameBuffer.
      fbp = new FrameBufferPanel(width, height);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 4");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    //jf.setLocationByPlatform(true);
      // Place the FrameBufferPanel in the JFrame.
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setVisible(true);

      // Register this object as the event listener for JFrame events.
      jf.addKeyListener(this);
      jf.addComponentListener(this);
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c)
      {
         Pipeline.debug = ! Pipeline.debug;
       //RasterizeAntialias.debug = ! RasterizeAntialias.debug;
      }
      else if ('/' == c)
      {
         currentModel = (currentModel + 1) % modelArray.size();
         scene.setModel( 0, modelArray.get(currentModel) );
      }
      else if ('?' == c)
      {
         currentModel = (currentModel - 1);
         if (currentModel < 0) currentModel = modelArray.size() - 1;
         scene.setModel( 0, modelArray.get(currentModel) );
      }
      else if ('a' == c)
      {
         RasterizeAntialias.doAntialiasing = ! RasterizeAntialias.doAntialiasing;
         System.out.print("Anti-aliasing is turned ");
         System.out.println(RasterizeAntialias.doAntialiasing ? "On" : "Off");
      }
      else if ('g' == c)
      {
         RasterizeAntialias.doGamma = ! RasterizeAntialias.doGamma;
         System.out.print("Gamma correction is turned ");
         System.out.println(RasterizeAntialias.doGamma ? "On" : "Off");
      }
      else if ('c' == c)
      {
         // Change the solid random color of the current model.
         ModelShading.setRandomColor(scene.getModel(0));
      }
      else if ('C' == c)
      {
         // Change each color in the current model to a random color.
         ModelShading.setRandomColors(scene.getModel(0));
      }
      else if ('e' == c && e.isAltDown())
      {
         // Change the random color of each vertex of the current model.
         ModelShading.setRandomVertexColors(scene.getModel(0));
      }
      else if ('e' == c)
      {
         // Change the solid random color of each edge of the current model.
         ModelShading.setRandomLineSegmentColors(scene.getModel(0));
      }
      else if ('E' == c)
      {
         // Change the random color of each end of each edge of the current model.
         ModelShading.setRainbowLineSegmentColors(scene.getModel(0));
      }
      else if ('m' == c) //display transformation information
      {
         displayTransformations = ! displayTransformations;
      }
      else if ('s' == c) // Scale all the models 10% smaller.
      {
         scaleModels(1.0/1.1);
      }
      else if ('S' == c) // Scale ALL the models 10% larger.
      {
         scaleModels(1.1);
      }
      else if ('x' == c) // Translate the model.
      {
         moveModels(-0.1, 0); // left
      }
      else if ('X' == c)
      {
         moveModels(+0.1, 0); // right
      }
      else if ('y' == c)
      {
         moveModels(0, -0.1); // down
      }
      else if ('Y' == c)
      {
         moveModels(0, +0.1);  // up
      }

      if (displayTransformations && ('m'==c
           ||'s'==c||'x'==c||'y'==c
           ||'S'==c||'X'==c||'Y'==c))
      {
         System.out.printf("deltaX = %.2f, deltaY = %.2f, " +
                           "deltaS = %.2f\n",
                            deltaX,        deltaY,
                            deltaS);
      }

      // Render again.
      final FrameBuffer fb = fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      fbp.update();
   }


   private void scaleModels(double deltaS)
   {
      this.deltaS *= deltaS;
      for (final Model m : modelArray)
      {
         for (int i = 0; i < m.vertexList.size(); ++i)
         {  // This does NOT work with perspective projection.
            final Vertex v = m.vertexList.get(i);
            m.vertexList.set(i, new Vertex(v.x * deltaS,
                                           v.y * deltaS,
                                           v.z));
         }
      }
   }


   private void moveModels(double deltaX, double deltaY)
   {
      this.deltaX += deltaX;
      this.deltaY += deltaY;
      for (final Model m : modelArray)
      {
         for (int i = 0; i < m.vertexList.size(); ++i)
         {
            final Vertex v = m.vertexList.get(i);
            m.vertexList.set(i, new Vertex(v.x + deltaX,
                                           v.y + deltaY,
                                           v.z));
         }
      }
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );

      //System.out.printf("JFrame [w = %d, h = %d]: " +
      //                  "FrameBufferPanel [w = %d, h = %d].\n",
      //                  jf.getWidth(), jf.getHeight(),
      //                  fbp.getWidth(), fbp.getHeight());

      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final FrameBuffer fb = new FrameBuffer(w, h);
      fbp.setFrameBuffer(fb);
      Pipeline.render(scene, fb);
      fbp.update();
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      print_help_message();

      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new InteractiveGRSModels_R5()
      );
   }


   private static void print_help_message()
   {
      System.out.println("Use the 'd/D' keys to toggle debugging information on and off for the current model.");
      System.out.println("Use the '/' key to cycle through the models.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the x/X, y/Y, z/Z, keys to translate the model along the x, y, z axes.");
      System.out.println("Use the s/S keys to scale the size of the model.");
      System.out.println("Use the 'm' key to toggle the display of transformation information.");
      System.out.println("Use the 'c' key to change the random solid model color.");
      System.out.println("Use the 'C' key to randomly change model's colors.");
    //System.out.println("Use the 'e' key to change the random vertex colors.");
      System.out.println("Use the 'e' key to change the random solid edge colors.");
      System.out.println("Use the 'E' key to change the random edge colors.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }
}
